﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;

using System.IO;
using System.Net;
using System.Threading;

namespace GPSFStileserver
{
    public partial class GPSFSts : Form
    {
        public GPSFSts()
        {
            InitializeComponent();
        }

        private void Form1_Load(object sender, EventArgs e)
        {
            System.Threading.ThreadPool.SetMaxThreads(50, 1000);
            System.Threading.ThreadPool.SetMinThreads(50, 50);
            _listener = new HttpListener();
            _listener.Prefixes.Add("http://+:8080/");

            appdir = Directory.GetCurrentDirectory() + "\\";
            
            tiledir = appdir + "tilesets\\";

            // build KML (each is unopenable dir pointing at 4 KMLs)
            string msg = "<kml>\n<Document>";
            msg = msg + "\n<name>GPSFStileserver</name><open>1</open>";
            // own region (coords)
            //msg = msg + "\n" + RegionXML(0, 0, 17, 1, true);

            
            dirs.Columns.Add("name", typeof(String));
            dirs.Columns.Add("ord", typeof(Int32));
            DirectoryInfo diri = new DirectoryInfo(tiledir);
            foreach (DirectoryInfo dirt in diri.GetDirectories())
            {
                DataRow row = dirs.NewRow();
                row["name"] = dirt.Name;
                dirs.Rows.Add(row);
            }
            Sport = Surl.Text + ":8080";
            NFO = "Serving at:\n\n" + Sport + "/index.kml\n";
            int count = 1;
            foreach (DataRow row in dirs.Select("", "name"))
            {
                row["ord"] = count;
                count++;
                msg = msg + DirectoryKML((string)row["name"]);
                NFO = NFO + "\n" + (string)row["name"];
            }

            msg = msg + "\n</Document>\n</kml>";
            INDEX = msg;


            // build NFO
            // >>
            //NFO = "There is a tileserver here, serving <LIST> at index.KML";
            NFOtxt.Text = NFO;
        }

        string Sport = "";

        private string DirectoryKML(string dir)
        {
            string msg = "";
            BinaryReader sr = new BinaryReader(new FileStream(tiledir + dir + "/GPSFS", FileMode.Open, FileAccess.Read, FileShare.Read));
            string gpsfsv = "";

            int TLx = -1;
            int TLy = -1;
            int TLzoom = 0;
            Int64 filetype = -1;
            int maxsize = -1;

            gpsfsv += sr.ReadChar();
            gpsfsv += sr.ReadChar();
            gpsfsv += sr.ReadChar();
            gpsfsv += sr.ReadChar();
            gpsfsv += sr.ReadChar();
            gpsfsv += sr.ReadChar();
            gpsfsv += sr.ReadChar();
            gpsfsv += sr.ReadChar();

            TLx = System.Convert.ToInt32(sr.ReadInt64());
            TLy = System.Convert.ToInt32(sr.ReadInt64());
            TLzoom = System.Convert.ToInt32(sr.ReadInt64());

            filetype = sr.ReadInt64();
            maxsize = System.Convert.ToInt32(sr.ReadInt64());

            sr.Close();

            int max_zoom = 1;
            int tiles = 0;
            while (tiles < maxsize)
            {
                tiles = System.Convert.ToInt32(TotalTiles(max_zoom));
                max_zoom *= 2;
                TLzoom++;
            }
            max_zoom /= 2;
            TLzoom--;

            int x = TLx / max_zoom;
            int y = TLy / max_zoom;
            int z = TLzoom;

            msg = msg + "\n<Folder>\n<name>" + dir + "</name>";
            msg = msg + "\n<Style><ListStyle><listItemType>checkHideChildren</listItemType></ListStyle></Style>";

            // KML points at 4 regions (coords) and KML's
            msg = msg + "\n" + NetworkLink(x, y, z, dir);
            msg = msg + "\n" + NetworkLink(x + 1, y, z, dir);
            msg = msg + "\n" + NetworkLink(x, y + 1, z, dir);
            msg = msg + "\n" + NetworkLink(x + 1, y + 1, z, dir);
            msg = msg + "\n</Folder>";
            return msg;
        }

        DataTable dirs = new DataTable();

        private string NFO = "";
        private string INDEX = "";

        private string appdir = "";
        private string tiledir = "";

        LatLonXYcalcs LLXY = new LatLonXYcalcs();

        HttpListener _listener;

        public void Start()
        {
            _listener.Start();
            while (StopServer.Checked == false)
            {
                //try
                //{
                    HttpListenerContext request = _listener.GetContext();
                    ThreadPool.QueueUserWorkItem(ProcessRequest, request);
                //}
                //catch (HttpListenerException) { break; }
                //catch (InvalidOperationException) { break; }
                //Application.DoEvents();
            }
            Stop();
            StopServer.Checked = false;
            StopServer.Enabled = false;

            StartServer.Enabled = true;
        }

        public void Stop()
        {
            _listener.Stop();
        }

        private void StartServer_Click(object sender, EventArgs e)
        {
            //MessageBox.Show("Note: to quit you will probably have to force quit the app.");

            StartServer.Enabled = false;
            StopServer.Enabled = true;
            Application.DoEvents();
            Start();
        }

        void ProcessRequest(object listenerContext)
        {
            //try
            //{
                var context = (HttpListenerContext)listenerContext;
                string req = context.Request.RawUrl; // get: /file.ext
                byte[] msg;

                if (req.EndsWith(".ico"))
                {
                    context.Response.StatusCode = (int)HttpStatusCode.NotFound;
                    msg = Encoding.UTF8.GetBytes("no");
                }
                else if (req == "/")
                {
                    context.Response.StatusCode = (int)HttpStatusCode.OK;
                    msg = Encoding.UTF8.GetBytes(NFO);
                }
                else if (req == "/index.kml")
                {
                    context.Response.StatusCode = (int)HttpStatusCode.OK;
                    msg = Encoding.UTF8.GetBytes(INDEX);
                }
                else if (req.EndsWith(".kml"))
                {
                    //try
                    //{


                    // >> overlay of different tileservers (gmap, bing, etc.)

                        int start = 1;
                        int end = req.IndexOf("/", start);
                        string dir = req.Substring(start, end - start);
                        start = end + 1;
                        end = req.IndexOf("/", start);
                        int zoom = Int32.Parse(req.Substring(start, end - start));
                        start = end + 1;
                        end = req.IndexOf("/", start);
                        int x = Int32.Parse(req.Substring(start, end - start));
                        start = end + 1;
                        end = req.IndexOf(".", start);
                        int y = Int32.Parse(req.Substring(start, end - start));
                        context.Response.StatusCode = (int)HttpStatusCode.OK;

                        string tile = Sport + req.Replace(".kml", "");

                        // >> but mostly redundant (only saves png send) {option to check tile in GPSFS, if not there tile = ""}

                        msg = Encoding.UTF8.GetBytes(kml(dir, zoom, x, y, tile));
                    //}
                    //catch
                    //{
                    //    context.Response.StatusCode = (int)HttpStatusCode.NotFound;
                    //    msg = Encoding.UTF8.GetBytes("<kml><Document><name>KML Error</name></Document></kml>"); // >> should never get here
                    //}
                }
                else
                {
                    //try
                    //{
                        int start = 1;
                        int end = req.IndexOf("/", start);
                        string dir = req.Substring(start, end - start);
                        start = end + 1;
                        end = req.IndexOf("/", start);
                        int zoom = Int32.Parse(req.Substring(start, end - start));
                        start = end + 1;
                        end = req.IndexOf("/", start);
                        int x = Int32.Parse(req.Substring(start, end - start));
                        start = end + 1;
                        end = req.Length;
                        int y = Int32.Parse(req.Substring(start, end - start));
                        context.Response.StatusCode = (int)HttpStatusCode.OK;

                        // get the tile and serve (this is finished)
                        // >> check if dir exists
                        // >> can add some debug (out of z up, out of z down, below xy bounds, empty/above xy bounds, served)
                        msg = tile(dir, zoom, x, y);
                    //}
                    //catch
                    //{
                    //    context.Response.StatusCode = (int)HttpStatusCode.NotFound;
                    //    msg = Encoding.UTF8.GetBytes("SERVE TRANSPARENT TILE<<<~~~$$$$"); // >> if get here may as well crash (due to IO)
                    //}
                }

                context.Response.ContentLength64 = msg.Length;
                try
                {
                    using (Stream s = context.Response.OutputStream)
                    {

                        s.Write(msg, 0, msg.Length);

                    }
                }
                catch
                {
                    // requester dropped
                }
            //}
            //catch (Exception ex) { Console.WriteLine("Request error: " + ex); }
        }

        private string kml(string dir, int z, int x, int y, string tile)
        {
            string msg = "<kml>\n<Document>\n";
            // own region (coords)
            //msg = msg + RegionXML(x, y, z, 1, false);

            // KML points at 4 regions (coords) and KML's
            msg = msg + "\n" + NetworkLink(x * 2, y * 2, z - 1, dir);
            msg = msg + "\n" + NetworkLink(x * 2 + 1, y * 2, z - 1, dir);
            msg = msg + "\n" + NetworkLink(x * 2, y * 2 + 1, z - 1, dir);
            msg = msg + "\n" + NetworkLink(x * 2 + 1, y * 2 + 1, z - 1, dir);

            // KML points at tile (coords) {note: draworder (increases as zoom in)}
            if (tile != "")
            {
                decimal[] Tcoord = LLXY.getLatLong(x, y, z); // top left
                decimal north = Tcoord[1];
                decimal west = Tcoord[0];
                decimal[] Bcoord = LLXY.getLatLong(x + 1, y + 1, z); // bottom right
                decimal south = Bcoord[1];
                decimal east = Bcoord[0];

                int dir_order = 0; // have sorted array of dir names and see which one (a = 0, b = 1, etc.)

                foreach (DataRow row in dirs.Select("name = '" + dir.Replace("%20", " ") + "'"))
                {
                    dir_order = (int)row["ord"];
                }

                // map supremacy: 100 * (100 - 1 - dir_order) + (27 - z)
                // zoom supremacy: 100 * (27 - z) + (100 - 1 - dir_order)

                msg = msg + "\n<GroundOverlay>\n";
                msg = msg + "<drawOrder>" + (100 * (27 - z) + (100 - 1 - dir_order)).ToString() + "</drawOrder>\n";
                msg = msg + "<Icon>\n";
                msg = msg + "<href>" + tile + "</href>\n";
                msg = msg + "</Icon>\n";
                msg = msg + "<LatLonBox>\n";
                msg = msg + "<north>" + north.ToString() + "</north><south>" + south.ToString() + "</south>\n";
                msg = msg + "<east>" + east.ToString() + "</east><west>" + west.ToString() + "</west>\n";
                msg = msg + "</LatLonBox>\n";
                msg = msg + "</GroundOverlay>";
            }
            msg = msg + "\n</Document>\n</kml>";

            return msg;
        }

        private string NetworkLink(int x, int y, int z, string dir)
        {
            decimal[] Tcoord = LLXY.getLatLong(x, y, z); // top left
            decimal north = Tcoord[1];
            decimal west = Tcoord[0];
            decimal[] Bcoord = LLXY.getLatLong(x + 1, y + 1, z); // bottom right
            decimal south = Bcoord[1];
            decimal east = Bcoord[0];

            int minLodPixels = scale_low;
            int maxLodPixels = scale_high;
            if (dir.ToLower().IndexOf("transp") == -1)
                maxLodPixels = -1;

            string msg = "<NetworkLink>\n";
            msg = msg + "<name>" + dir + "/" + z + "/" + x + "/" + y + ".kml" + "</name>\n";
            msg = msg + "<Region>\n";
            msg = msg + "<LatLonAltBox>\n";
            msg = msg + "<north>" + north.ToString() + "</north><south>" + south.ToString() + "</south>\n";
            msg = msg + "<east>" + east.ToString() + "</east><west>" + west.ToString() + "</west>\n";
            msg = msg + "<minAltitude>0.000000</minAltitude><maxAltitude>0.000000</maxAltitude>\n";
            msg = msg + "</LatLonAltBox>\n";
            msg = msg + "<Lod>\n";
            msg = msg + "<minLodPixels>" + minLodPixels.ToString() + "</minLodPixels><maxLodPixels>" + maxLodPixels.ToString() + "</maxLodPixels>\n";
            msg = msg + "<minFadeExtent>0</minFadeExtent><maxFadeExtent>0</maxFadeExtent>\n";
            msg = msg + "</Lod>\n";
            msg = msg + "</Region>\n";
            msg = msg + "<Link>\n";
            msg = msg + "<href>" + Sport + "/" + dir + "/" + z + "/" + x + "/" + y + ".kml</href>\n";
            msg = msg + "<viewRefreshMode>onRegion</viewRefreshMode>\n";
            msg = msg + "</Link>\n";
            msg = msg + "</NetworkLink>";
            return msg;
        }

        int scale_low = 200; // 128; // 225
        int scale_high = 400; // 512; // -1

        private byte[] tile(string dir, int zoom, int x, int y)
        {
            dir = dir.Replace("%20", " ");
            // >> write a nice 

            byte[] msg;

            int IDXfile = 0;
            BinaryReader sr = new BinaryReader(new FileStream(tiledir + dir + "/GPSFS", FileMode.Open, FileAccess.Read, FileShare.Read));
            string gpsfsv = "";

                int TLx = -1;
                int TLy = -1;
                int TLzoom = 0;
                Int64 filetype = -1;
                int maxsize = -1;

                gpsfsv += sr.ReadChar();
                gpsfsv += sr.ReadChar();
                gpsfsv += sr.ReadChar();
                gpsfsv += sr.ReadChar();
                gpsfsv += sr.ReadChar();
                gpsfsv += sr.ReadChar();
                gpsfsv += sr.ReadChar();
                gpsfsv += sr.ReadChar();

                TLx = System.Convert.ToInt32(sr.ReadInt64());
                TLy = System.Convert.ToInt32(sr.ReadInt64());
                TLzoom = System.Convert.ToInt32(sr.ReadInt64());

                filetype = sr.ReadInt64();
                maxsize = System.Convert.ToInt32(sr.ReadInt64());

                byte[] IDXba = sr.ReadBytes(8 * (System.Convert.ToInt32(maxsize) + 1));

            if (zoom < TLzoom)
            {
                sr.Close();
                return File.ReadAllBytes(appdir + "transparent.png");
            }
            
            int max_zoom = 1;
            int gz = 1;
            int tiles = 0;
            while (tiles < maxsize)
            {
                tiles = System.Convert.ToInt32(TotalTiles(max_zoom));
                max_zoom *= 2;
                if (zoom > TLzoom)
                {
                    TLx /= 2;
                    TLy /= 2;
                    gz *= 2;
                }
                TLzoom++;
            }
            max_zoom /= 2;
            TLzoom--;

            if (zoom > TLzoom)
            {
                sr.Close();
                return File.ReadAllBytes(appdir + "transparent.png");
            }

            // offset
            x -= TLx;
            y -= TLy;

            if (x < 0 || y < 0)
            {
                sr.Close();
                return File.ReadAllBytes(appdir + "transparent.png");
            }

            int[] fol = TileFileOffsetLength(gz, x, y, max_zoom, IDXba, dir, maxsize);
            if (fol[2] > 0)
            {
                if (IDXfile != fol[0]) // (int)tile["IDXfile"])
                {
                    IDXfile = fol[0]; // (int)tile["IDXfile"];
                    sr.Close();
                    string file_suff = IDXfile.ToString();
                    if (file_suff == "0")
                        file_suff = "";
                    sr = new BinaryReader(new FileStream(tiledir + dir + "/GPSFS" + file_suff, FileMode.Open, FileAccess.Read, FileShare.Read));
                }
                // seek
                sr.BaseStream.Position = fol[1]; // (int)tile["IDXoffset"];
                // read
                msg = sr.ReadBytes(fol[2]); // (int)tile["length"])));
                
            }
            else
            {
                sr.Close();
                return File.ReadAllBytes(appdir + "transparent.png");
            }

            sr.Close();
            return msg;
        }

        public int TotalTiles(int top_dir)
        {
            int maxsize = 0;

            int zoom = 1;
            int span = 2;

            while (zoom <= top_dir)
            {
                maxsize += span * span;
                zoom *= 2;
                span *= 2;
            }

            return maxsize;
        }

        private int[] TileFileOffsetLength(int zoom, int x, int y, int max_zoom, byte[] IDXba, string dir, int max_size)
        {
            int tile = 0;

            int file = -1;
            int offset = -1;
            int length = 0;

            int s_zoom = 1;
            while (s_zoom < zoom)
            {
                tile += System.Convert.ToInt32(Math.Pow((max_zoom / s_zoom) * 2, 2));
                s_zoom *= 2;
            }

            int span = 2 * max_zoom / s_zoom;

            if (x < span && y < span)
            {
                tile += span * y;
                tile += x;

                if (tile < max_size)
                {
                    file = System.BitConverter.ToInt32(IDXba, tile * 8);
                    offset = System.BitConverter.ToInt32(IDXba, tile * 8 + 4);
                    if (offset >= 0)
                    {
                        int offset1 = System.BitConverter.ToInt32(IDXba, (tile + 1) * 8 + 4);
                        if (offset1 > 0)
                        {
                            length = offset1 - offset;
                        }
                        else if (System.BitConverter.ToInt32(IDXba, (tile + 1) * 8 + 4) < 0)
                        {
                            length = offset1 * -1;
                        }
                        else // if (offset1 == 0)
                        {
                            string file_suff = file.ToString();
                            if (file_suff == "0")
                                file_suff = "";
                            length = System.Convert.ToInt32(new FileInfo(tiledir + dir + "/GPSFS" + file_suff).Length) - offset;
                        }
                    }
                }
            }

            return (new int[] { file, offset, length });
        }

/*if (!File.Exists(path))
{
    context.Response.StatusCode = (int)HttpStatusCode.NotFound;
    msg = Encoding.UTF8.GetBytes("Sorry, that page does not exist");
}
else
{
    context.Response.StatusCode = (int)HttpStatusCode.OK;
    msg = File.ReadAllBytes(path);
}*/

/*
scan tilesets directory (in own dir) for subdirs
check subdirs for GPSFS
- will serve these with name of subdir (wo_)

serve: /index.kml
create (to serve) index.kml {as string}
dirs (can't open), each pointing at top 4 tiles and kml's

requests: index.kml, .kml, else tiles

GUI:
/index.kml (click link)
table (serving): names of subdir


serve tile: /subdir/zoom/x/y
get zoom, x, y {try}
if zoom not in range serve transparent
if zoom in range serve from GPSFS (transparent if empty)


serve kml: /subdir/zoom/x/y.kml
get zoom, x, y {try}
serve KML pointing at 4 tiles and kml's
{optional - check if tiles pointing at exist [checkbox]}
*/

        // >> threading
    }
}
